from __future__ import print_function

import logging
import sys

try:
    import parasail
except ImportError as e:
    logging.error("Could not load parasail library. Please try reinstalling "
                  "parasail using pip.")
    sys.exit(1)

import operator

from qcat import adapters, calibration
from qcat import config
from qcat.utils import revcomp


if parasail.can_use_sse2():
    parasail_sg_stat = parasail.sg_stats_striped_32
    parasail_sg = parasail.sg_striped_32
else:
    logging.warning("Warning: SSE not supported, falling back to standard alignment")
    parasail_sg_stat = parasail.sg_stats
    parasail_sg = parasail.sg


def extract_barcode_region(read_sequence, adapter_template, barcode_set_index,
                           alignment_stop_ref, qcat_config):
    """
    Extracts the region from the read that is expected to contain
    the barcode sequence.
    :param read_sequence: Sequence of the read
    :type read_sequence: str
    :param adapter_template: Adapter layout with masked barcode region
    :type adapter_template: AdapterLayout
    :param barcode_set_index:
    :param alignment_stop_ref: Position on the read that the last bp of the
    adapter template is aligned to
    :type alignment_stop_ref: int
    :param qcat_config: qcatConfig object
    :type qcat_config: qcatConfig
    :return: Part of the read that will be used to identify the barcode
    :rtype str
    """
    adapter_length = adapter_template.get_adapter_length()
    barcode_end = adapter_template.get_barcode_end(barcode_set_index)
    barcode_length = adapter_template.get_barcode_length(barcode_set_index)

    barcode_end_ref = alignment_stop_ref - (adapter_length - barcode_end) + 1
    barcode_start_ref = barcode_end_ref - barcode_length

    # Extend region to allow for indels
    barcode_start_ref -= min(qcat_config.extracted_barcode_extension,
                             barcode_start_ref)
    barcode_end_ref += min(qcat_config.extracted_barcode_extension,
                           len(read_sequence) - barcode_end_ref)

    return read_sequence[barcode_start_ref:barcode_end_ref + 1]


def find_highest_scoring_barcode(barcode_region_read,
                                 barcode_set,
                                 qcat_config,
                                 upstream_context="",
                                 downstream_context="",
                                 compute_identity=False):
    """
    Aligns all the barcodes from barcode_set to the barcode region that
    was extracted from the read and chooses the barcode with the best alignment
    score. Also computes a quality score for the barcode.
    Quality score = (best - secondbest) / best
    Where best is the highest alignment score found in all barcode
    alignments and secondbest is the second best score. Range is 0-1,
    a reasonable cutoff is between 0.1 and 0.2.

    :param barcode_region_read: Sequence extracted from the read that is
    supposed to contain the barcode
    :type barcode_region_read: str
    :param barcode_set: List of possible Barcode tuples
    :type barcode_set: List
    :param upstream_context: Sequence upstream of barcode in adapter_template
    :type upstream_context: str
    :param downstream_context: Sequence downstream of barcode in
    adapter_template
    :type downstream_context: str
    :param qcat_config: qcatConfig object
    :type qcat_config: qcatConfig
    :return: Best barcode and quality score
    :rtype Barcode, float
    """
    max_score = None
    max_identity = 0.0
    max_end = -1
    second_best_score = None
    second_best_identity = 0.0
    max_barcode = None
    q_score = 0

    if not barcode_region_read:
        return max_barcode, q_score, max_identity, max_end

    for barcode in barcode_set:

        if compute_identity:
            align = parasail_sg_stat
        else:
            align = parasail_sg

        aligned_barcode = align(s1=barcode_region_read,
                                s2=upstream_context +
                                   barcode.sequence +
                                   downstream_context,
                                open=1,
                                extend=1,
                                matrix=qcat_config.matrix_barcode)

        score = aligned_barcode.score * 100.0 / (1.0 * len(upstream_context + barcode.sequence + downstream_context))

        identity = 0.0
        if compute_identity:
            identity = float(aligned_barcode.matches) / len(barcode.sequence)

        if not max_score or max_score < score:
            second_best_score = max_score
            second_best_identity = max_identity
            max_score = score
            max_barcode = barcode
            max_identity = identity
            max_end = aligned_barcode.end_query
        elif not second_best_score or second_best_score < score:
            second_best_score = score
            second_best_identity = identity

    # print(max_score, file=sys.stderr)
    # if max_score > 0:
    #     q_score = int(float(max_score - second_best_score) / float(max_score) * 100.0)
    q_score = max_score

    return max_barcode, q_score, max_score, max_end


def align_adapter_identity(adapter_sequence, adapter_length, read_sequence,
                           barcode_length, qcat_config):
    """
    Aligns a single adapter template to the read an computes the
    identity for the alignment

    :param adapter_sequence: Sequence of adapter template
    :type adapter_sequence: str
    :param adapter_length: Length of adapter sequence
    :type adapter_length: int
    :param read_sequence: Read sequence
    :type read_sequence: str
    :param barcode_length: Length of the barcode. Number of
    bp will be ignored for identity computation
    :type barcode_length: int
    :param qcat_config: qcatConfig object
    :type qcat_config: qcatConfig
    :return: Parasail alignment object (None if partial match),
    alignment identity (0.0 if partial match)
    :rtype Result, int
    """
    if not read_sequence or not adapter_sequence:
        return None, 0.0

    aligned_adapter = parasail_sg_stat(s1=read_sequence,
                                       s2=adapter_sequence,
                                       open=qcat_config.gap_open,
                                       extend=qcat_config.gap_extend,
                                       matrix=qcat_config.matrix)

    # Check whether the whole adapter is aligned to the read or only a suffix
    partial_match = aligned_adapter.length < (adapter_length * 0.85)
    # partial_match = False
    if partial_match:
        # Currently partial matches between the read and the reference
        # are discarded
        aligned_adapter = None
        adapter_identity = 0.0
    else:
        # Compute alignment identity but do not take into account
        # the barcode region which is masked by Ns
        adapter_identity = float(aligned_adapter.matches) / \
                           float(aligned_adapter.length - barcode_length)

    return aligned_adapter, adapter_identity


def align_adapter(adapter_sequence, read_sequence, qcat_config):
    """
    Aligns a single adapter template to the read an computes the
    identity for the alignment

    :param adapter_sequence: Sequence of adapter template
    :type adapter_sequence: str
    :param adapter_length: Length of adapter sequence
    :type adapter_length: int
    :param read_sequence: Read sequence
    :type read_sequence: str
    :param barcode_length: Length of the barcode. Number of
    bp will be ignored for identity computation
    :type barcode_length: int
    :param qcat_config: qcatConfig object
    :type qcat_config: qcatConfig
    :return: Parasail alignment object (None if partial match),
    alignment identity (0.0 if partial match)
    :rtype Result, int
    """
    if not read_sequence or not adapter_sequence:
        return None, 0.0

    aligned_adapter = parasail_sg(s1=read_sequence,
                                  s2=adapter_sequence,
                                  open=qcat_config.gap_open,
                                  extend=qcat_config.gap_extend,
                                  matrix=qcat_config.matrix)

    return aligned_adapter, 0.0


def extract_align_sequence(read_sequence, rev_comp, length):
    """
    Extracts the part of the read that will be scanned for an adapter

    :param read_sequence: str
    :param length: Length of sequence that should be scanned. Starting at the 5'
    end of the read. If 0, full read is returned
    :type length: int
    :return: part read sequence
    :rtype str
    """
    if read_sequence:
        align_sequence = read_sequence
    else:
        align_sequence = ""
    if length > 0:
        if not rev_comp:
            align_sequence = align_sequence[:length]
        else:
            align_sequence = revcomp(align_sequence[-length:])

    return align_sequence


def compute_adapter_identity(adapter_template, read_sequence, qcat_config):
    aligned_adapter, adapter_identity = \
        align_adapter_identity(adapter_template.get_adapter_sequences(),
                               adapter_template.get_adapter_length(),
                               read_sequence,
                               adapter_template.get_barcode_length(
                                   0) + adapter_template.get_barcode_length(1),
                               qcat_config)
    return adapter_identity


def eval_adapter_template(adapter_template, read_sequence,
                          qcat_config, identity=True):
    """
    Extracts the adapter sequence with masked barcode from the adapter_template
    and aligns (semi-global) it to the 5' end, 3' end or full read.

    :param adapter_template: Adapter layout object
    :type adapter_template: AdapterLayout
    :param read_sequence: Read sequence
    :type read_sequence: str
    :param qcat_config: qcatConfig object
    :type qcat_config: qcatConfig
    :return: Position the last bp of the adapter is aligned to in the read,
    Identity of adapter alignment, Score of adapter alignment
    :rtype: int, float, float
    """
    adapter_end_position = -1
    adapter_score = -1

    # Align adapter to extracted sequence
    if identity:
        ret = align_adapter_identity(adapter_template.get_adapter_sequences(),
                                     adapter_template.get_adapter_length(),
                                     read_sequence,
                                     adapter_template.get_barcode_length(0) +
                                     adapter_template.get_barcode_length(1),
                                     qcat_config)
        aligned_adapter, adapter_identity = ret
    else:
        ret = align_adapter(adapter_template.get_adapter_sequences(),
                            read_sequence,
                            qcat_config)
        aligned_adapter, adapter_identity = ret

    if aligned_adapter is not None:
        adapter_end_position = aligned_adapter.end_query
        adapter_score = aligned_adapter.score

    return adapter_end_position, adapter_identity, adapter_score


def get_norm_socre(template, score, qcat_config):
    """
    Normalize alignment score

    :param template: Barcoding adapter
    :param score: Alignment score
    :param qcat_config: Config
    :return: Normalized score
    """
    bc_len = template.get_barcode_length(0) + template.get_barcode_length(1)
    a_len = template.get_adapter_length()
    return score * 100.0 / ((a_len - bc_len) * qcat_config.match + bc_len * qcat_config.nmatch)


def find_best_adapter_template(adapter_templates, read_sequence,
                               qcat_config):
    """
    Aligns all passed adapter templates to the read sequence returns the one
    with the highest alignment score

    :param adapter_templates: List of AdapterLayout objects
    :type adapter_templates: List
    :param read_sequence: Sequence of the read
    :type read_sequence: str
    :param qcat_config: qcatConfig object
    :type qcat_config: qcatConfig
    :return: Sequence of best adapter, alignment identity, last position of
    the aligned adapter in the read sequence,
    last position of barcode in the adapter template, length of the barcode
    :rtype: str, float, int, int, int
    """
    best_adapter_score = -1.0
    best_adapter_end_position = -1
    best_adapter_template = -1

    if not adapter_templates or not read_sequence:
        return best_adapter_template, best_adapter_end_position, best_adapter_score

    # If single adapter template is passed, convert it to list
    if not isinstance(adapter_templates, list):
        adapter_templates = [adapter_templates]

    for i, template in enumerate(adapter_templates):

        if not template.get_adapter_sequences():
            continue

        ret = eval_adapter_template(adapter_template=template,
                                    read_sequence=read_sequence,
                                    qcat_config=qcat_config,
                                    identity=False)
        adapter_end_position, _, adapter_score = ret

        adapter_score = get_norm_socre(template, adapter_score, qcat_config)

        if best_adapter_score < adapter_score:
            best_adapter_score = adapter_score
            best_adapter_template = i
            best_adapter_end_position = adapter_end_position

    return best_adapter_template, best_adapter_end_position, best_adapter_score


def build_return_dict(best_barcode, best_barcode_score,
                      best_adapter,
                      best_adapter_end,
                      exit_status,
                      trim5p=0, trim3p=0
                      ):
    """
    Builds dict containing all return values

    :param best_barcode: Barcode with the best alignment score
    :type  best_barcode: Barcode
    :param best_barcode_score: Normalized alignment score
    :type  best_barcode_score: float
    :param best_adapter: Adapter identified
    :type  best_adapter: AdapterLayout
    :param best_adapter_end: End position of adapter on the read
    :type  best_adapter_end: int
    :return: Dictionary
    """
    return_dict = {"barcode":                best_barcode,
                   "barcode_score":          best_barcode_score,
                   "adapter":                best_adapter,
                   "adapter_end":            best_adapter_end,
                   "trim5p":                 trim5p,
                   "trim3p":                 trim3p,
                   "exit_status":            exit_status
                  }

    return return_dict


def empty_return_dict():
    """
    Return empty Barcoding result dict

    :return: Empty barcode result dict
    """
    return build_return_dict(
        best_barcode=None,
        best_barcode_score=0.0,
        best_adapter=None,
        best_adapter_end=0,
        exit_status=1,
        trim3p=0,
        trim5p=0
    )


class BarcodeScanner(object):
    """
    Abstract base calss for BarcodeScanners
    """

    def __init__(self, min_quality, kit_name, kit_folder=None,
                 enable_filter_barcodes=False, scan_middle_adapter=False):
        """
        Init

        :param min_quality: Minimum barcode qulity
        :param kit_name: Name of the kit to call barcodes for
        :param kit_folder: Folder containing kit config files
        :param enable_filter_barcodes: Remove low abundance adapters in batch
        mode
        :param scan_middle_adapter: Scann full read for adapters
        """

        available_kits = adapters.populate_adapter_layouts(kit_folder)

        # Parameters
        self.min_quality = min_quality

        self.layouts = []
        self.override_kit_name = None

        self.enable_filter_barcodes = enable_filter_barcodes
        self.scan_middle_adapter = scan_middle_adapter

        # Get kets
        if kit_name and kit_name.lower() != 'auto':
            for layout in available_kits:
                if kit_name.lower() == layout.kit.lower():
                    self.layouts.append(layout)
        else:
            for layout in available_kits:
                if layout.auto_detect:
                    self.layouts.append(layout)

    @staticmethod
    def get_name():
        """
        Get name of barcode scanner algorithm

        :return: Name
        """
        raise NotImplemented("Abstract class")

    def barcode_count(self):
        """
        Number of supported barcodes

        :return: barcode number
        """
        raise NotImplemented("Abstract class")

    def scan(self, read_sequence, read_qualities, barcoding_kits,
             non_barocding_kits, qcat_config=config.qcatConfig()):
        """

        :param read_sequence:
        :param read_qualities:
        :param barcoding_kits:
        :param non_barocding_kits:
        :param qcat_config:
        :return:
        """
        raise NotImplemented("Abstract class")

    def scan_middle(self, sequence, kit_name, qcat_config):

        detected_adapters = self.get_adapters(kit_name)

        middle_barcode = self.scan(sequence[qcat_config.max_align_length:-qcat_config.max_align_length],
                                   None,
                                   detected_adapters,
                                   [],
                                   qcat_config=qcat_config)

        if not middle_barcode or middle_barcode[
            'barcode_score'] < 50.0:

            middle_barcode = self.scan(revcomp(sequence[
                                       qcat_config.max_align_length:-qcat_config.max_align_length]),
                                       None,
                                       detected_adapters,
                                       [],
                                       qcat_config=qcat_config)
            if not middle_barcode or middle_barcode[
                'barcode_score'] < 50.0:
                return False
            else:
                logging.debug(
                    "middle adapter ({}) found on rev strand with score {} at {}".format(
                        kit_name,
                        middle_barcode[
                            'barcode_score'],
                        middle_barcode[
                            'adapter_end']),
                    file=sys.stderr)
                return True
        else:
            logging.debug(
                "middle adapter ({}) found with score {} at {}".format(kit_name,
                                                                       middle_barcode[
                                                                           'barcode_score'],
                                                                       middle_barcode[
                                                                           'adapter_end']),
                file=sys.stderr)
            return True

    def detect_barcode(self,
                       read_sequence,
                       read_qualities=None,
                       qcat_config=config.qcatConfig()):

        if not self.override_kit_name:
            kits = self.layouts
        else:
            kits = self.get_adapters(self.override_kit_name)

        # Check 5' end
        align_seq_5p = extract_align_sequence(read_sequence,
                                              False,
                                              qcat_config.max_align_length)

        barcode_dict_5p = self.scan(align_seq_5p,
                                    read_qualities,
                                    kits,
                                    [],
                                    qcat_config=qcat_config)

        if barcode_dict_5p and barcode_dict_5p[
            'barcode_score'] < self.min_quality:
            barcode_dict_5p = empty_return_dict()

        trim_5p = 0
        if barcode_dict_5p['adapter_end'] > 0:
            trim_5p = barcode_dict_5p['adapter_end']

        # Check 3' end
        align_seq_5p_rc = extract_align_sequence(read_sequence,
                                                 True,
                                                 qcat_config.max_align_length)

        barcode_dict_5p_rc = self.scan(align_seq_5p_rc,
                                       read_qualities,
                                       kits,
                                       [],
                                       qcat_config=qcat_config)

        if barcode_dict_5p_rc and \
                barcode_dict_5p_rc['barcode_score'] < self.min_quality:
            barcode_dict_5p_rc = empty_return_dict()

        trim_3p = len(read_sequence)
        if barcode_dict_5p_rc['adapter'] and barcode_dict_5p_rc[
            'adapter_end'] > 0:
            trim_3p = trim_3p - barcode_dict_5p_rc['adapter_end']

        # Find best barcode
        results = [barcode_dict_5p, barcode_dict_5p_rc]

        best = None
        best_score = 0.0
        for i in range(0, len(results)):
            result = results[i]
            if result:
                if result['barcode_score'] > best_score:
                    best_score = result['barcode_score']
                    best = result

        if not best:
            best = empty_return_dict()
        else:
            min_score = 60
            if barcode_dict_5p['barcode'] and barcode_dict_5p_rc['barcode']:
                if barcode_dict_5p['barcode_score'] >= min_score and \
                        barcode_dict_5p_rc['barcode_score'] >= min_score and \
                        barcode_dict_5p['barcode'].id != barcode_dict_5p_rc['barcode'].id:
                    best = empty_return_dict()
                    best['exit_status'] = 1002

        if self.scan_middle_adapter and best['adapter'] and self.scan_middle(read_sequence, best['adapter'].kit, qcat_config):
            best = empty_return_dict()
            best['exit_status'] = 997

        best["trim5p"] = trim_5p
        best["trim3p"] = trim_3p

        if best["trim3p"] < best["trim5p"]:
            # Only happens for reads that only consist of the barcode
            best["trim5p"] = 0

        return best

    def get_adapters(self, kit_name):
        result = []
        for layout in self.layouts:
            if kit_name.lower() == layout.kit.lower():
                result.append(layout)
        return result

    def get_adapter(self, kit_name):
        for layout in self.layouts:
            if kit_name.lower() == layout.kit.lower():
                return layout

    def scan_end(self, sequence, reverse, qcat_config):
        # Check 5' end
        align_seq_5p = extract_align_sequence(sequence,
                                              reverse,
                                              qcat_config.max_align_length)

        ret = find_best_adapter_template(adapter_templates=self.layouts,
                                         read_sequence=align_seq_5p,
                                         qcat_config=qcat_config)

        best_adapter_template_index, aligned_adapter_end, best_adapter_score = ret

        return self.layouts[best_adapter_template_index], aligned_adapter_end, best_adapter_score

    def scan_ends(self, read_sequence, qcat_config):
        # 5' end
        adapter_5p, end5p, score_5p = self.scan_end(read_sequence, False, qcat_config)
        # 3' end
        adapter_3p, end3p, score_3p = self.scan_end(read_sequence, True, qcat_config)
        end3p = qcat_config.max_align_length - end3p

        if score_5p > score_3p:
            return adapter_5p, adapter_3p
        else:
            return adapter_3p, adapter_5p

    @staticmethod
    def update_kit_count(adapter, adapter_counts):
        if adapter:
            kit = adapter.kit
            if kit not in adapter_counts.keys():
                adapter_counts[kit] = 0
            adapter_counts[kit] += 1
        else:
            if "none" not in adapter_counts.keys():
                adapter_counts["none"] = 0
            adapter_counts["none"] += 1

    @staticmethod
    def get_most_abundant_kits(adapter_counts):
        if not adapter_counts:
            return None
        return sorted(adapter_counts.items(), key=operator.itemgetter(1), reverse=True)[0][0]

    def detect_kit(self, read_sequences, qcat_config):

        adapter_counts = {}

        adapter_ends = []

        for read_sequence in read_sequences:
            adapter_1, adapter_2 = self.scan_ends(read_sequence, qcat_config)
            self.update_kit_count(adapter_1, adapter_counts)
            # self.update_kit_count(adapter_2, adapter_counts)
            # self.update_kit_count(adapter, adapter_counts)

            # adapter_ends.append((end5p, end3p))

        kit_name = self.get_most_abundant_kits(adapter_counts)

        return kit_name, adapter_ends

    @staticmethod
    def update_barcode_count(result, barcode_count):
        if result and result['barcode']:
            if result['barcode'].id not in barcode_count:
                barcode_count[result['barcode'].id] = 0
            barcode_count[result['barcode'].id] += 1
        else:
            if "0" not in barcode_count:
                barcode_count["0"] = 0
            barcode_count["0"] += 1

    @staticmethod
    def get_valid(barcode_counts, min_perc=0.20):
        max_barcode_count = 0
        for n in barcode_counts.values():
            max_barcode_count = max(max_barcode_count, n)

        min_count = int(max_barcode_count * min_perc)

        valid_barcodes = []
        for bc, count in barcode_counts.items():
            if count > min_count:
                valid_barcodes.append(bc)

        return valid_barcodes

    def filter_barcodes(self, barcode_count, results):
        valid_barcodes = self.get_valid(barcode_count, 0.05)

        for i, barcode in enumerate(results):
            if barcode and barcode['barcode'] and barcode['barcode'].id not in valid_barcodes:
                results[i] = empty_return_dict()
        return results

    def detect_barcode_batch(self, read_sequences, read_qualities=[None],
                             qcat_config=config.qcatConfig()):
        # barcode_count = [0] * 1000
        barcode_count = {}

        kit_name, _ = self.detect_kit(read_sequences, qcat_config)
        results = []

        self.override_kit_name = kit_name
        for read_sequence, read_quality in zip(read_sequences, read_qualities):
            result = self.detect_barcode(read_sequence, read_quality, qcat_config)
            self.update_barcode_count(result, barcode_count)
            results.append(result)

        self.override_kit_name = None

        if self.enable_filter_barcodes:
            results = self.filter_barcodes(barcode_count, results)

        return results



import os

import pkg_resources

import numpy as np

from qcat import config
from qcat.scanner import BarcodeScanner
from qcat.scanner_base import empty_return_dict, build_return_dict

try:
    from brill.model import load_model
    from brill import seq_util
    from brill import encoding
    use_brill = True
except ImportError as e:
    use_brill = False


class BarcodeScannerBrill(BarcodeScanner):

    def __init__(self, min_quality=None, kit_folder=None, kit=None, enable_filter_barcodes=False, scan_middle_adapter=False):

        if min_quality is None:
            min_quality = 10

        super(BarcodeScannerBrill, self).__init__(min_quality,
                                                  kit,
                                                  kit_folder=kit_folder,
                                                  enable_filter_barcodes=enable_filter_barcodes,
                                                  scan_middle_adapter=scan_middle_adapter)

        os.environ['TF_CPP_MIN_LOG_LEVEL'] = '2'

        # Load all available brill models
        self.brill_models = {}
        for layout in self.layouts:
            ret = self.load_model(layout)
            if ret:
                self.brill_models[layout.kit] = ret

    def load_model(self, adapter):
        brill_model = self.find_model(adapter.model)
        if not brill_model:
            return None

        # Load model:
        return load_model(brill_model), adapter.model_len

    @staticmethod
    def get_name():
        return "brill"

    @staticmethod
    def run_brill(model, model_len, read_sequence, end5p=None, end3p=None):
        if not end5p:
            end5p = model_len
        if not end3p:
            end3p = model_len

        x = encoding.build_predictor_matrix(read_sequence[end5p - model_len:end5p],
                                            read_sequence[-end3p:][:model_len])

        prediction = model.predict_proba(x, verbose=0)

        return BarcodeScannerBrill.process_prediction(prediction)

    @staticmethod
    def find_model(path):
        if not path:
            return None
        if os.path.exists(path):
            return path
        else:
            resource_model = pkg_resources.resource_filename(__name__, os.path.join("resources/models", path))
            if os.path.exists(resource_model):
                return resource_model
            else:
                raise RuntimeError("Couldn't locate model {}".format(path))

    def detect_barcode(self,
                       read_sequence,
                       read_qualities=None,
                       qcat_config=config.qcatConfig()):

        result = empty_return_dict()

        if not self.override_kit_name:
            detected_adapter, _ = self.scan_ends(read_sequence, qcat_config)
            override_kit_name = detected_adapter.kit
        else:
            override_kit_name = self.override_kit_name

        if override_kit_name in self.brill_models:
            brill_model, brill_model_len = self.brill_models[override_kit_name]

            if len(read_sequence) > brill_model_len:
                bc, qscore = self.run_brill(brill_model, brill_model_len, read_sequence)

                barcode = None
                adapter = None
                if bc and bc != 'none':
                    adapter = self.get_adapter(override_kit_name)
                    barcode = adapter.get_barcode_set()[int(bc) - 1]

                if bc and qscore >= self.min_quality:
                    result = build_return_dict(
                        best_barcode=barcode,
                        best_barcode_score=qscore,
                        best_adapter=adapter,
                        best_adapter_end=90,
                        trim5p=90,
                        trim3p=90,
                        exit_status=0
                    )

        return result

    @staticmethod
    def process_prediction(p):
        """ Process CNN prediction, return barcode and q-score.

        :param p: CNN prediction.
        :param min_qual: Minimum quality for classified barcodes.
        :return: barcode and qscore
        :rtype: tuple

        """
        # Select winning barcode:
        bc = np.argmax(p[0])
        # Calculate error probability and q-score:
        prob_error = 1 - p[0][bc]
        qscore = seq_util.prob_to_phred(prob_error, max_q=200)

        if bc == 0:
            # Unclassified bin predicted:
            bc = None
        return bc, qscore





from __future__ import print_function

import logging

try:
    import pandas as pd
except ImportError as e:
    logging.warning("Could not load pandas. To use qcat-roc please install pandas.")
import sys

from argparse import ArgumentParser, RawDescriptionHelpFormatter

from qcat import cli


def parse_args(argv):
    """
    Commandline parser

    :param argv: Command line arguments
    :type argv: List
    :param defaults: qcat config object that holds default values passed to
    ArgumentParser
    :type defaults: qcatConfig
    :return: None
    """
    usage = "Command line interface to barcode detection library"
    parser = ArgumentParser(description=usage,
                            formatter_class=RawDescriptionHelpFormatter)
    parser.add_argument(dest="FASTQ", nargs="?", default=None)
    # parser.add_argument("-t", "--tsv", dest="TSV", default=None, required=False)
    # parser.add_argument("-n", "--name", dest="NAME", default="-", required=False)
    # parser.add_argument("-d", "--dataset", dest="DATASET", default="-", required=False)
    # parser.add_argument("-s", "--summary", action="store_true", dest="SUMMARY", required=False)
    # parser.add_argument("-g", "--guppy_summary", dest="GUPPY", default=None, required=False)
    args = parser.parse_args(argv)

    return args


def summary(truebc_col, bc_col, score_col, min_score):
    counts = {'total': 0, 'correct': 0, 'incorrect': 0, 'true_negative': 0,
              'false_negative': 0, 'false_positive': 0, 'score': min_score}
    for truebc, bc, score in zip(truebc_col, bc_col, score_col):
        if score < min_score:
            bc = "none"

        truebc = [str(truebc)]
        if bc == "none":
            if truebc == ["none"] or truebc == "none":
                counts["true_negative"] += 1
            else:
                counts["false_negative"] += 1
        else:
            if truebc == ["none"] or truebc == "none":
                counts["false_positive"] += 1
            else:
                if str(bc) in truebc:
                    counts['correct'] += 1
                else:
                    counts['incorrect'] += 1
        counts['total'] += 1

    return counts


def create_roc(input_file):
    data = pd.read_csv(input_file, sep="\t")

    col_names = ['total', 'correct', 'incorrect', 'true_negative',
                 'false_negative', 'false_positive', 'score', 'program',
                 'dataset']
    my_df = pd.DataFrame(columns=col_names)

    for q in range(0, 1001):
        counts = summary(data['truebc'], data['bc'], data['score'], q / 10.0)

        counts['program'] = data['program'][0]
        counts['dataset'] = data['dataset'][0]
        my_df = my_df.append(counts, ignore_index=True)
    return my_df


def main(argv=sys.argv[1:]):
    """
    Basic command line interface to qcat.

    :param argv: Command line arguments
    :type argv: list
    :return: None
    :rtype: NoneType
    """
    args = parse_args(argv=argv)

    if args.FASTQ:
        df = create_roc(args.FASTQ)
    else:
        df = create_roc(sys.stdin)

    df.to_csv("/dev/stdout", index=False, sep='\t')
    # print(df.to_string(index=False))

if __name__ == '__main__':

    main()



import re

from collections import namedtuple

BarcodePosition = namedtuple("BarcodePosition", "start end length")


class AdapterLayout:
    """
    Represents the layout of adapter. Usually one kit has a specific
    adapter layout
    """

    def __init__(self, kit, sequence, barcode_set_1, barcode_set_2, description, auto_detect=False, model=None, model_len=None, name=None):
        """
        Init

        :param name: Name of adapter/kit
        :param layout_id: Unique ID
        :param sequence: Descriptive name for the adapter
        :param barcode_set_1: Set of barcodes supported by the kit
        :param barcode_set_2: Set of secondary barcodes supported (double barconding)
        :param kits: List of kits using this adapter
        :param description: Free text description
        """
        self.kit = kit
        self.auto_detect = auto_detect
        self.model = model
        self.model_len = model_len
        if not name:
            self.name = self.kit
        else:
            self.name = name

        self.sequence = sequence.upper()
        if not self.sequence or re.findall("[^ATGCNX]", self.sequence):
            raise RuntimeError("Invalid adapter sequence: {}".
                               format(self.sequence))

        self.barcode_set_1 = barcode_set_1
        self.barcode_set_2 = barcode_set_2
        self.barcode_count = 0

        for barcode_set in [self.barcode_set_1, self.barcode_set_2]:
            if barcode_set:
                self.barcode_count += 1

        self.barcode_pos_1 = BarcodePosition(-1, -1, 0)
        self.barcode_pos_2 = BarcodePosition(-1, -1, 0)

        if self.barcode_set_1:
            self.barcode_pos_1 = self.get_placeholder_pos(self.sequence, 0)
            for barcode in self.barcode_set_1:
                if len(barcode.sequence) != self.barcode_pos_1.length:
                    raise RuntimeError("Adapter length does not match "
                                       "place holder length: {}, {}".
                                       format(len(barcode.sequence),
                                              self.barcode_pos_1.length))

        if self.barcode_set_2:
            self.barcode_pos_2 = self.get_placeholder_pos(self.sequence, 1)
            for barcode in self.barcode_set_2:
                if len(barcode.sequence) != self.barcode_pos_2.length:
                    raise RuntimeError("Adapter length does not match "
                                       "place holder length: {}, {}".
                                       format(len(barcode.sequence),
                                              self.barcode_pos_2.length))

        self.description = description

    @staticmethod
    def get_placeholder_pos(adapter_template, index=0):
        """
        Returns start and end position of the placehoder
        region for the barcode in the adapter

        :param adapter_template: adapter sequence with Ns
        instead of barcode sequence
        :return: Start and end position of barcode placeholder
        in adapter template
        :rtype: int, int
        """
        start = -1
        end = -1
        length = 0

        matches = list(re.finditer('N+', adapter_template))

        if matches:
            if len(matches) > index:
                start = matches[index].start(0)
                end = matches[index].end(0) - 1
                length = end - start + 1

        return BarcodePosition(start, end, length)

    def __repr__(self):
        return {"Kit": self.kit,
                "Description": self.description}.__repr__()

    def get_barcode_end(self, index=0):
        """
        Get the position of the last barcode bp in the adapter

        :param index: 0 for single barcoding, 0 or 1 for double barcoding
        :return:
        """
        if index == 0:
            return self.barcode_pos_1.end
        elif index == 1:
            return self.barcode_pos_2.end
        else:
            raise RuntimeError("Invalid barcode index: {}. Must be 0 or 1 "
                               "(for double barcoding)".format(index))

    def get_barcode_length(self, index=0):
        """
        Length of barcode in template. Usually the same for all kits

        :param index: 0 for single barcoding, 0 or 1 for double barcoding
        :return:
        """
        if index == 0:
            return self.barcode_pos_1.length
        elif index == 1:
            return self.barcode_pos_2.length
        else:
            raise RuntimeError("Invalid barcode index: {}. Must be 0 or 1 "
                               "(for double barcoding)".format(index))

    def get_adapter_sequences(self, barcode_seq=None):
        """
        Returns the sequence of the adapter. If barcode_seq is omitted,
        the region containing the barcode will be masked with Ns.

        :param barcode_seq: sequence of the barcode
        :return: str
        """
        if barcode_seq:
            seq_5p = self.sequence[:self.barcode_pos_1.start]
            seq_3p = self.sequence[self.barcode_pos_1.end + 1:]
            return seq_5p + barcode_seq + seq_3p
        else:
            return self.sequence

    def get_full_adapter_sequences(self, context=None):
        """
        Returns all possible adapter sequences including barcodes

        :param context:
        :return: Barcode, Adapter sequence
        :rtype: Barcode, List
        """
        if self.barcode_count == 0:
            yield None, self.sequence
        elif self.barcode_count == 1:
            seq_5p = self.sequence[:self.barcode_pos_1.start]
            seq_3p = self.sequence[self.barcode_pos_1.end + 1:]
            for barcode in self.barcode_set_1:
                full_sequence = seq_5p + barcode.sequence + seq_3p
                if context:
                    full_sequence = seq_5p[
                                    :-context] + barcode.sequence + seq_3p[
                                                                    context:]
                yield barcode, full_sequence

    def get_adapter_length(self):
        """
        Length of full adapter sequence including barcode

        :return: int
        """
        return len(self.get_adapter_sequences())

    def get_barcode_set(self, index=0):
        """
        Get a list of barcodes supported by the kit

        :param index: 0 for single barcoding, 0 or 1 for double barcoding
        :return:
        """
        if index == 0:
            return self.barcode_set_1
        elif index == 1:
            return self.barcode_set_2
        else:
            raise RuntimeError("Invalid barcode index: {}. Must be 0 or 1 "
                               "(for double barcoding)".format(index))

    def get_upstream_context(self, n, index=0):
        """
        Return n bp upstream of the barcode

        :param index: 0 for single barcoding, 0 or 1 for double barcoding
        :type n: int
        :param n: Number of bp
        :type n: int
        :return:
        """
        if index == 0:
            barcode = self.barcode_pos_1
        elif index == 1:
            barcode = self.barcode_pos_2
        else:
            raise RuntimeError("Invalid barcode index: {}. Must be 0 or 1 "
                               "(for double barcoding)".format(index))

        if barcode.end > -1:
            upstream_start = max(0, barcode.start - n)
            return self.sequence[upstream_start:barcode.start]
        else:
            return ""

    def get_downstream_context(self, n, index=0):
        """
        Return n bp downstream of the barcode

        :param index: 0 for single barcoding, 0 or 1 for double barcoding
        :type n: int
        :param n: Number of bp
        :type n: int
        :return:
        """
        if index == 0:
            barcode = self.barcode_pos_1
        elif index == 1:
            barcode = self.barcode_pos_2
        else:
            raise RuntimeError("Invalid barcode index: {}. Must be 0 or 1 "
                               "(for double barcoding)".format(index))

        if barcode.end > -1:
            downstream_end = min(self.get_adapter_length(),
                                 barcode.end + n + 1)
            return self.sequence[barcode.end + 1:downstream_end]
        else:
            return ""

    def is_doulbe_barcode(self):
        """
        Returns true if the kit has two barcodes per adapter
        end of the read

        :return: if kit has two barcodes per adapter
        :rtype: bool
        """
        return self.barcode_set_2 is not None


from __future__ import print_function

import logging

try:
    import pysam
except ImportError as e:
    logging.warning(
        "Could not load pysam. To use qcat-eval-full please install pysam.")
import math
import sys
try:
    import mappy as mp
except ImportError as e:
    logging.warning(
        "Could not load pysam. To use qcat-eval-full please install mappy.")

import os
import glob

from argparse import ArgumentParser, RawDescriptionHelpFormatter


from qcat import config
from qcat.scanner import BarcodeScanner

LOOKUP = []


def qstring_to_phred(quality):
    """ Compute standard phred scores from a quality string. """
    qscores = []
    if quality is not None:
        qscores = [ord(q) - 33 for q in quality]
    return qscores


for q in range(100):
    LOOKUP.append(pow(10, -.1 * q))


def _compute_mean_qscore(scores):
    """ Returns the phred score corresponding to the mean of the probabilities
    associated with the phred scores provided.

    :param scores: Iterable of phred scores.

    :returns: Phred score corresponding to the average error rate, as
        estimated from the input phred scores.
    """
    if not scores:
        return 0.0
    sum_prob = 0.0
    for val in scores:
        sum_prob += LOOKUP[val]
    mean_prob = sum_prob / len(scores)
    return -10.0 * math.log10(mean_prob)



def parse_args(argv):
    """
    Commandline parser

    :param argv: Command line arguments
    :type argv: List
    :param defaults: qcat config object that holds default values passed to
    ArgumentParser
    :type defaults: qcatConfig
    :return: None
    """
    usage = "Command line interface to barcode detection library"
    parser = ArgumentParser(description=usage,
                            formatter_class=RawDescriptionHelpFormatter)
    parser.add_argument(dest="FASTQ", nargs="?", default=None)
    parser.add_argument("-t", "--tsv", dest="TSV", required=True)
    parser.add_argument("-n", "--name", dest="NAME", required=True)
    parser.add_argument("-g", "--genomes", dest="GENOMES", required=True)
    parser.add_argument("--max", dest="MAX", default=-1, required=False)
    # parser.add_argument("-s", "--summary", action="store_true", dest="SUMMARY", required=False)
    # parser.add_argument("-g", "--guppy_summary", dest="GUPPY", default=None, required=False)
    args = parser.parse_args(argv)

    return args


def _parse_reads_info(comment):
    """
    Parse ONT FASTQ header

    :param comment: str
    :return: dict
    """
    single_read_info = {'runid': 'unknown'}
    if comment:
        for pair in comment.split(" "):
            cols = pair.split('=')
            if len(cols) == 2:
                single_read_info[cols[0]] = cols[1]
    return single_read_info


def create_reference(mapping, reference_file, fasta_path):
    ref_count = 0
    with open(reference_file, "w") as ref:
        for g in glob.glob(fasta_path):
            with pysam.FastxFile(g) as fh:
                id = os.path.splitext(os.path.basename(g))[0]
                for entry in fh:
                    if id in mapping:
                        ref_count += 1
                        name = ','.join([str(x) for x in mapping[id]])

                        print(">" + str(mapping[id][0]) + " " + entry.name,
                              entry.comment, file=ref)
                        print(entry.sequence, file=ref)
    if ref_count != len(mapping.keys()):
        raise RuntimeError(
            "Couldn't find all references, please check mappings!")
    return ref_count


def get_truebc(a, sequence, min_aln_perc=0.85, adapter_len=200):
    #             print(bc, score)
    hits = list(a.map(sequence))

    truebc = None
    truebc_assignment = None
    if len(hits) == 0:
        truebc_assignment = "Not_found"
    else:
        if len(hits) == 1:
            hit = hits[0]
            if hit.blen > min_aln_perc * (len(sequence) - adapter_len):
                truebc = hit.ctg
                truebc_assignment = "Unique"
            else:
                truebc = hit.ctg
                truebc_assignment = "Partial"
        else:
            #             for hit in hits:
            #                 print(hit)

            truebcs = set([hit.ctg for hit in hits])
            truebc = ",".join(truebcs)
            if len(truebcs) == 1:
                truebc_assignment = "Fragmented"
            else:
                truebc = "Multiple"
                truebc_assignment = "Multiple"
    #             print(truebc, truebc_assignment)

    #             print(truebc, truebc_assignment)
    #             print("---")

    return truebc, truebc_assignment


def main(argv=sys.argv[1:]):
    """
    Basic command line interface to qcat.

    :param argv: Command line arguments
    :type argv: list
    :return: None
    :rtype: NoneType
    """
    args = parse_args(argv=argv)

    reference_file = args.GENOMES

    a = mp.Aligner(reference_file, preset="map-ont")  # load or build index

    if not a: raise Exception("ERROR: failed to load/build index")

    output_file = open(args.TSV, "w")

    detector = BarcodeScanner.factory(kit="NBD10X", min_quality=0)

    print("dataset", "id", "length", "qscore", "score", "barcode", "truebc",
          "truebc_assignment", "middle_adapter", "non_matching_bc", "sequence",
          sep='\t', file=output_file)
    name = args.NAME
    f = args.FASTQ
    i = 0
    print(name, f)
    with pysam.FastxFile(f) as fh:
        for entry in fh:
            i += 1

            info = _parse_reads_info(entry.comment)
            if 'truebc' in info:
                truebc = info['truebc']
                truebc_assignment = "Pre_assigned"
            else:
                truebc, truebc_assignment = get_truebc(a, entry.sequence)

            c = config.get_default_config()
            result = detector.detect_barcode(entry.sequence,
                                             qcat_config=c)

            middle_adapter = detector.scan_middle(entry.sequence, "NBD10X",
                                                  c)

            q_score = _compute_mean_qscore(qstring_to_phred(entry.quality))
            bc = "none"
            score = 0.0
            if result['barcode']:
                bc = result['barcode'].id
                score = result['barcode_score']

            no_seq = False
            seq = "-"
            if not no_seq:
                seq = entry.sequence

            print(name, entry.name, len(entry.sequence), q_score, score, bc,
                  truebc, truebc_assignment, middle_adapter,
                  result['exit_status'] == 1002, seq,
                  sep='\t', file=output_file)

            #             if i == 10000:
            #                 break
            if i % 1000 == 0:
                print(".", end="")
            if i % 50000 == 0:
                print("")

            if i == int(args.MAX):
                print("Reach maximumg, stopping.")
                break
    print("")

    output_file.close()

    print("finished")


if __name__ == '__main__':

    main()



try:
    from string import maketrans
except ImportError:
    maketrans = str.maketrans

LOOKUP = []

for q in range(100):
    LOOKUP.append(pow(10,-.1*q))


def qstring_to_phred(quality):
    """ Compute standard phred scores from a quality string. """
    qscores = []
    if quality is not None:
        qscores = [ord(q) - 33 for q in quality]
    return qscores


def revcomp(seq):
    return seq.translate(maketrans('ACGTacgtRYMKrymkVBHDvbhd', 'TGCAtgcaYRKMyrkmBVDHbvdh'))[::-1]


def mean_error_prob(scores):
    """ Returns the phred score corresponding to the mean of the probabilities
    associated with the phred scores provided.

    :param scores: Iterable of phred scores.

    :returns: Phred score corresponding to the average error rate, as
        estimated from the input phred scores.
    """
    if not scores:
        return -1.0
    sum_prob = 0.0
    for val in scores:
        sum_prob += LOOKUP[val]
    mean_prob = sum_prob / len(scores)
    # return -10.0 * log10(mean_prob)
    return mean_prob





from __future__ import print_function

import logging
import time
import sys
import os
import six

from Bio.SeqIO.FastaIO import SimpleFastaParser
from Bio.SeqIO.QualityIO import FastqGeneralIterator

from argparse import ArgumentParser, RawDescriptionHelpFormatter, ArgumentTypeError

from qcat import __version__, adapters, config
from qcat import scanner
from qcat.adapters import Barcode
from qcat.scanner import get_modes, factory, get_kits_info, get_kits


def get_mode(args):
    if args.MODE_GUPPY:
        logging.warning("Guppy/albacore barcoding is not yet supported. "
                        "Falling back to epi2me.")
        return "epi2me"
    if args.MODE_EPI2ME:
        return "epi2me"
    if args.MODE_SIMPLE:
        return "simple"
    if args.MODE_DUAL:
        return "dual"

    return "epi2me"


def check_minqual_arg(x):
    x = float(x)
    if x < 0.0 or x > 100.0:
        raise ArgumentTypeError("Minimum quality must be a value between 0 and 100.")
    return x


def check_kit_arg(x):
    x = str(x)
    if x.lower() == "dual":
        raise ArgumentTypeError(
            "-k dual and --kit dual are not supported any more. "
            "Please use --dual instead.")
    return x


def parse_args(argv):
    """
    Commandline parser

    :param argv: Command line arguments
    :type argv: List
    :param defaults: qcat config object that holds default values passed to
    ArgumentParser
    :type defaults: qcatConfig
    :return: None
    """
    defaults = config.get_default_config()
    usage = "Python command-line tool for demultiplexing Oxford Nanopore reads from FASTQ files"
    parser = ArgumentParser(description=usage,
                            formatter_class=RawDescriptionHelpFormatter)
    parser.add_argument("-V", '--version',
                        action='version',
                        version='%(prog)s ' + __version__)
    parser.add_argument('-l', "--log",
                        dest="log",
                        type=str,
                        default="INFO",
                        help="Print debug information")
    parser.add_argument("--quiet",
                        dest="QUIET",
                        action='store_true',
                        help="Don't print summary")

    general_group = parser.add_argument_group('General settings')
    general_group.add_argument("-f", "--fastq",
                        type=str,
                        dest="fastq",
                        help="Barcoded read file")
    general_group.add_argument('-b', "--barcode_dir",
                        dest="barcode_dir",
                        type=str,
                        default=None,
                        help="If specified, qcat will demultiplex reads "
                             "to this folder")
    general_group.add_argument('-o', "--output",
                        dest="output",
                        type=str,
                        default=None,
                        help="Output file trimmed reads will be written to "
                             "(default: stdout).")
    general_group.add_argument('-q', "--min-quality",
                        dest="min_qual",
                        type=check_minqual_arg,
                        default=None,
                        help="Minimum barcode quality. "
                             "Must be between 0 and 100. (default: 60)")
    general_group.add_argument("--detect-middle",
                        dest="DETECT_MIDDLE",
                        action='store_true',
                        help="Search for adapters in the whole read")
    general_group.add_argument('-t', "--threads",
                        dest="threads",
                        type=int,
                        default=1,
                        help="Number of threads. Only works with in guppy mode")
    general_group.add_argument("--tsv",
                        dest="tsv",
                        action='store_true',
                        help="Prints a tsv file containing barcode information "
                             "each read to stdout.")
    general_group.add_argument("--trim",
                        dest="TRIM",
                        action='store_true',
                        help="Remove adapter and barcode sequences from reads.")
    general_group.add_argument("-k", "--kit",
                               dest="kit",
                               choices=get_kits(),
                               type=check_kit_arg,
                               default="auto",
                               help="Sequencing kit. Specifying the correct kit "
                                    "will improve sensitivity and specificity and "
                                    "runtime (default: auto)")
    general_group.add_argument("--list-kits",
                               dest="list_kits",
                               action='store_true',
                               help="List all supported kits")

    # Mode
    mode_group = parser.add_argument_group("Demultiplexing modes")
    group = mode_group.add_mutually_exclusive_group()

    group.add_argument("--guppy",
                        dest="MODE_GUPPY",
                        action = 'store_true',
                        help="Use Guppy's demultiplexing algorithm (default: false)")
    group.add_argument("--epi2me",
                       dest="MODE_EPI2ME",
                       action='store_true',
                       help="Use EPI2ME's demultiplexing algorithm (default: true)")
    group.add_argument("--dual",
                       dest="MODE_DUAL",
                       action='store_true',
                       help="Use dual barcoding algorithm")
    group.add_argument("--simple",
                       dest="MODE_SIMPLE",
                       action='store_true',
                       help="Use simple demultiplexing algorithm. Only looks for "
                            "barcodes, not for adapter sequences. Use only for testing "
                            "purposes!")

    # EPI2ME
    epi2me_group = parser.add_argument_group('EPI2ME options (only valid with --epi2me)')
    epi2me_group.add_argument("--no-batch",
                              dest="nobatch",
                              action='store_true',
                              help="Don't use information from multiple reads for kit detection (default: false)")
    epi2me_group.add_argument("--filter-barcodes",
                              dest="FILTER_BARCODES",
                              action='store_true',
                              help="Filter rare barcode calls when run in batch mode")

    simple_group = parser.add_argument_group('Simple options (only valid with --simple)')
    simple_group.add_argument("--simple-barcodes",
                              dest="SIMPLE_BARCODES",
                              choices=['standard', 'extended'],
                              default="standard",
                              help="Use 12 (standard) or 96 (extended) barcodes for demultiplexing")

    # parser.add_argument("--adapter-sequences",
    #                     type=str,
    #                     dest="adapter_yaml",
    #                     default=None,
    #                     help="YAML file or folder containing YAML files that"
    #                          "describe the Adapter sequences.")

    # parser.add_argument('--no-header',
    #                     action='store_true',
    #                     dest='no_header',
    #                     help="Do not write header to output tsv file."
    #                          "Useful if results from different runs should be"
    #                          "combined")

    args = parser.parse_args(argv)

    return args


def extract_fastx_comment(header):
    """
    Split FASTA/Q header into name and comment

    :param header: FASTA/Q header line
    :return: name, comment
    :rtype: tuple
    """
    cols = header.replace("\t", " ").split(" ")
    name = cols[0]
    comment = None
    if len(cols) > 1:
        comment = " ".join(cols[1:])

    return name, comment


def is_fastq(filename):
    if not filename:
        logging.debug("Reading from stdin. Assuming FASTQ format. "
                     "Please use -f/--fasta to run with FASTA files.")
        return True
    c = None
    with open(filename) as f:
        c = f.read(1)

    if c == "@":
        return True
    elif c == ">":
        return False
    else:
        raise ValueError("Invalid input file. File must start with "
                           "'@' or '>'. Current file starts with: " + c)


def iter_fastx(reads_fx, fastq, batchsize):
    """
    Return iterator for FASTA/Q file

    :param reads_fx: filename of FASTX file
    :return: None
    """
    # batch = []

    names = []
    comments = []
    seqs = []
    quals = []

    # fastq = is_fastq(reads_fx)
    if fastq:

        if reads_fx:
            logging.debug("Reading from file {}".format(reads_fx))
            f = open(reads_fx)
        else:
            logging.debug("Reading from stdin")
            f = sys.stdin

        try:
            for title, seq, qual in FastqGeneralIterator(f):
                name, comment = extract_fastx_comment(title)

                # batch.append((name, comment, seq, qual))
                names.append(name)
                comments.append(comment)
                seqs.append(seq)
                quals.append(qual)

                if len(names) >= batchsize:
                    yield names, comments, seqs, quals
                    # batch = []
                    names = []
                    comments = []
                    seqs = []
                    quals = []
        except ValueError as e:
            logging.error(e.message)
            sys.exit(1)

        if reads_fx:
            logging.debug("Closing file {}".format(reads_fx))

                # yield name, comment, seq, qual

    else:
        with open(reads_fx) as f:
            for title, seq in SimpleFastaParser(f):
                name, comment = extract_fastx_comment(title)

                # batch.append((name, comment, seq, None))
                names.append(name)
                comments.append(comment)
                seqs.append(seq)
                quals.append(None)

                if len(names) >= batchsize:
                    yield names, comments, seqs, quals
                    # batch = []
                    names = []
                    comments = []
                    seqs = []
                    quals = []

    if len(names) > 0:
        # yield batch
        yield names, comments, seqs, quals


def get_output_file(output_files, out_folder, barcode_dict, fastq):

    bc_id = "none"
    if barcode_dict['barcode']:
        bc_id = barcode_dict['barcode'].name

    if bc_id not in output_files.keys():
        if fastq:
            output_files[bc_id] = open(
                os.path.join(out_folder, bc_id + ".fastq"), "w")
        else:
            output_files[bc_id] = open(
                os.path.join(out_folder, bc_id + ".fasta"), "w")
    return output_files[bc_id]


def write_to_file(trimmed_output_file, output_files, out_folder,
                  name, comment, sequence, quality, fastq, result):

    if not comment:
        comment = ""
    if not quality:
        quality = ""

    if out_folder:
        out_file = get_output_file(output_files, out_folder, result, fastq)
        if fastq:
            print("@" + name + " " + comment, sequence, "+", quality, sep="\n",
                  file=out_file)
        else:
            print(">" + name + " " + comment, sequence, sep="\n",
                  file=out_file)

    # if trimmed_output_file:
    else:

        bc = "none"
        if result["barcode"]:
            bc = str(result['barcode'].id)

        comment = "{} barcode={}".format(comment, bc)

        if fastq:
            print("@" + name + " " + comment, sequence, "+", quality, sep="\n",
                  file=trimmed_output_file)
        else:
            print(">" + name + " " + comment, sequence, sep="\n",
                  file=trimmed_output_file)


def close_files(output_files):
    for f in output_files.values():
        f.close()


def adapter_found(adapter_dist, adapter):
    adapterid = "none"
    if adapter:
        adapterid = adapter.kit

    if adapterid not in adapter_dist.keys():
        adapter_dist[adapterid] = 0
    adapter_dist[adapterid] += 1


def barcode_found(barcode_dist, barcode):
    bcid = "none"
    if barcode:
        bcid = barcode.name

    if bcid not in barcode_dist.keys():
        barcode_dist[bcid] = 0
    barcode_dist[bcid] += 1


def print_barcode_hist(barcode_dist, adapter_dist, total_reads):
    adapters_detected = 0
    for key, value in six.iteritems(adapter_dist):
        if key != "none":
            adapters_detected += value
    logging.info("Adapters detected in %d of %d reads" % (adapters_detected, total_reads))
    for adapterid in sorted(adapter_dist):
        perc = adapter_dist[adapterid] * 100.0 / total_reads
        logging.info("%15s %6d: | %20s | %6s %%" %(adapterid, adapter_dist[adapterid],
                                     int(perc / 5) * "#", "{:.2f}".format(perc)))

    barcodes_detected = 0
    for key, value in six.iteritems(barcode_dist):
        if key != "none":
            barcodes_detected += value
    logging.info("Barcodes detected in %d of %d adapters" % (barcodes_detected, total_reads))
    for bcid in sorted(barcode_dist):
        perc = barcode_dist[bcid] * 100.0 / total_reads
        logging.info("%15s %6d: | %20s | %6s %%" % (bcid, barcode_dist[bcid],
                                    int(perc / 5) * "#", "{:.2f}".format(perc)))


def write_multiplexing_result(barcode_dict, comment, name, sequence, tsv):
    """
    Writes results to TSV (--tsv) file.

    :param barcode_dict: detected barcode
    :param comment: FASTQ comment
    :param name:  FASTQ name
    :param sequence: FASTQ sequence
    :param tsv: TSV path
    :return: None
    """
    if barcode_dict['barcode']:
        kit_name = None
        if barcode_dict['adapter']:
            kit_name = barcode_dict['adapter'].kit

        if tsv:
            print(name,
                  len(sequence),
                  barcode_dict['barcode'].id,
                  barcode_dict['barcode_score'],
                  kit_name,
                  barcode_dict['adapter_end'],
                  comment,
                  sep="\t")
    else:
        if tsv:
            print(name,
                  len(sequence),
                  "none",
                  "-1",
                  "none",
                  "-1",
                  comment,
                  sep='\t')


def qcat_cli(reads_fq, kit, mode, nobatch, out,
               min_qual, tsv, output, threads, trim, adapter_yaml, quiet, filter_barcodes, middle_adapter,
               qcat_config):
    """
    Runs barcode detection for each read in the fastq file
    and print the read name + the barcode to a tsv file

    :param reads_fq: Path to fastq file
    :type reads_fq: str
    :param no_header: Print header or not for output tsv file
    :type no_header: bool
    :param kit: Which kit was used for sequencing
    :type kit: str
    :param simple: Triggers "simple" detection mode where only
    :type simple: bool
    barcodes will be aligned to the read. Adapter detection is skipped.
    :param nobatch: Switch of "batch" mode (First the kit and the barcodes used
    in the sample are detected from a small subset of read. Then, only these
    adapter/barcodes will be used for barcode detection)
    :type nobatch: bool
    :param qcat_config: qcatConfig object
    :type qcat_config: qcatConfig
    :return: None
    """

    total_reads = 0
    barcode_dist = {}
    adapter_dist = {}

    notrimming = not trim

    detector = factory(mode=mode,
                       kit=kit,
                       min_quality=min_qual,
                       kit_folder=adapter_yaml,
                       enable_filter_barcodes=filter_barcodes,
                       scan_middle_adapter=middle_adapter,
                       threads=threads)

    if tsv:
        print("name", "length", "barcode", "score", "kit",
              "adapter_end", "comment", sep="\t")

    fastq = is_fastq(reads_fq)
    output_files = {}

    trimmed_output_file = sys.stdout
    if output:
        trimmed_output_file = open(output, "w")

    if out:
        if not os.path.exists(out):
            os.makedirs(out)

    batch_size = 4000
    if nobatch:
        batch_size = 1

    for names, comments, seqs, quals in iter_fastx(reads_fq, fastq, batch_size):
        # Detect adapter/barcode
        if nobatch:
            results = [detector.detect_barcode(read_sequence=seqs[0],
                                               read_qualities=quals[0],
                                               qcat_config=qcat_config)]
        else:
            results = detector.detect_barcode_batch(read_sequences=seqs,
                                                    read_qualities=quals,
                                                    qcat_config=qcat_config)
        for name, comment, sequence, quality, result in zip(names,
                                                            comments,
                                                            seqs,
                                                            quals,
                                                            results):
            total_reads += 1
            # Record which adapter/barcode was found
            barcode_found(barcode_dist, result['barcode'])
            adapter_found(adapter_dist, result['adapter'])

            if not notrimming:
                trim_5p = result["trim5p"]
                trim_3p = result["trim3p"]
                sequence = sequence[trim_5p:trim_3p]
                if quality:
                    quality = quality[trim_5p:trim_3p]

            # Write tsv result file
            write_multiplexing_result(result,
                                      comment,
                                      name,
                                      sequence,
                                      tsv)
            # Write FASTQ/A files
            if not tsv:
                write_to_file(trimmed_output_file,
                              output_files,
                              out,
                              name,
                              comment,
                              sequence,
                              quality,
                              fastq,
                              result)

    if out:
        close_files(output_files)

    if not quiet:
        print_barcode_hist(barcode_dist, adapter_dist, total_reads)

    if output:
        trimmed_output_file.close()


def barcodes_from_fasta(filename):
    barcodes = []
    for names, comments, seqs, quals in iter_fastx(filename, False, 1):
        for name, comment, sequence in zip(names, comments, seqs):
            barcodes.append(Barcode("BC{}".format(name),
                                    int(name),
                                    sequence,
                                    True))

    return barcodes


def main(argv=sys.argv[1:]):
    """
    Basic command line interface to qcat.

    :param argv: Command line arguments
    :type argv: list
    :return: None
    :rtype: NoneType
    """
    try:
        args = parse_args(argv=argv)

        qcat_config = config.get_default_config()

        numeric_level = getattr(logging, args.log.upper(), None)
        if not isinstance(numeric_level, int):
            raise ValueError('Invalid log level: %s' % args.log.upper())
        logging.basicConfig(level=numeric_level, format='%(message)s')

        if args.list_kits:
            kits = get_kits_info()
            for kit in sorted(kits.keys()):
                if kit != "auto" and kit != "DUAL":
                    logging.info("{:<30}{}".format(kit, kits[kit]))
            return

        mode = get_mode(args)
        kit = args.kit
        if mode == "simple":
            kit = args.SIMPLE_BARCODES

        start = time.time()
        qcat_cli(reads_fq=args.fastq,
                 # no_header=args.no_header,
                 kit=kit,
                 mode=mode,
                 nobatch=args.nobatch,
                 out=args.barcode_dir,
                 min_qual=args.min_qual,
                 tsv=args.tsv,
                 output=args.output,
                 threads=args.threads,
                 trim=args.TRIM,
                 adapter_yaml=None,#=args.adapter_yaml, #args.barcode_fa,
                 quiet=args.QUIET,
                 filter_barcodes=args.FILTER_BARCODES,
                 middle_adapter=args.DETECT_MIDDLE,
                 qcat_config=qcat_config)
        end = time.time()

        if not args.QUIET:
            logging.info("Demultiplexing finished in {0:.2f}s".format(end - start))
    except FileNotFoundError as e:
        logging.error(e)
    except ValueError as e:
        logging.error(e)


if __name__ == '__main__':

    main()



import logging

from qcat import config
from qcat.adapters import get_barcodes_simple
from qcat.scanner_base import BarcodeScanner, find_highest_scoring_barcode, \
    build_return_dict, empty_return_dict


class BarcodeScannerSimple(BarcodeScanner):

    def __init__(self, min_quality=None, kit_folder=None, kit=None, enable_filter_barcodes=False, scan_middle_adapter=False, threads=1):

        if min_quality is None:
            min_quality = 60

        if threads != 1:
            logging.warning("Multi threading is not yet supported in "
                            "simple mode. Falling back to using a "
                            "single thread.")

        super(BarcodeScannerSimple, self).__init__(min_quality,
                                                   None,
                                                   kit_folder=kit_folder,
                                                   enable_filter_barcodes=enable_filter_barcodes,
                                                   scan_middle_adapter=scan_middle_adapter
                                                   )
        self.barcodes = get_barcodes_simple(kit)

    @staticmethod
    def get_name():
        return "simple"

    def barcode_count(self):
        if self.barcodes:
            return len(self.barcodes) + 1
        else:
            return super(BarcodeScannerSimple, self).barcode_count()

    def scan(self,
             read_sequence,
             read_qualities,
             bc_adapter_templates,
             nobc_adapter_templates,
             qcat_config=config.qcatConfig()):
        """
            Aligns all passed barcodes to the read sequence and chooses the one
            with the highest alignment score. Simplest version of barcode detection
            that is independent of the sequencing kit but has higher FN, FP and
            incorrect rate for barcode assignments. Currently does not implement any
            identity thresholds. Therefor, filtering by quality score is required.

            :param read_sequence: Sequence of the read
            :type read_sequence: str
            :param read_qualities: Base qualities of the read in Sanger encoding
            :type read_qualities: str
            :param barcode_set: List of possible Barcode tuples
            :type barcode_set: List
            :param qcat_config: qcatConfig object
            :type qcat_config: qcatConfig
            :return: see build_return_dict
            :rtype: Dictionary
            """

        exit_status = 0

        best_barcode, best_barcode_q_score, identity, barcode_end = \
            find_highest_scoring_barcode(barcode_region_read=read_sequence,
                                         barcode_set=self.barcodes,
                                         qcat_config=qcat_config,
                                         compute_identity=True)

        # barcode_err_probe = 0.0
        # if best_barcode:
        #     barcode_err_probe = compute_adapter_error_prob(read_qualities,
        #                                                    barcode_end,
        #                                                    len(
        #                                                        best_barcode.sequence))

        # If identity is too low or error prob too high, report as unclassified
        if identity < self.min_quality:
            return empty_return_dict()

        return build_return_dict(
            best_barcode=best_barcode,
            best_barcode_score=best_barcode_q_score,
            best_adapter=None,
            best_adapter_end=barcode_end,
            exit_status=exit_status
        )


import glob
import os
import logging
import pkg_resources
import yaml
from collections import namedtuple

from qcat.layout import AdapterLayout

Barcode = namedtuple("Barcode", "name id sequence fwd_strand")


def barcode2yaml(bc):
    barcode = {}
    barcode['name'] = bc.name
    barcode['id'] = bc.id
    barcode['sequence'] = bc.sequence
    barcode['fwd_strand'] = bc.fwd_strand
    return barcode


def barcodes2yaml(barcodes):
    barcode_set = []
    for bc in barcodes or []:
        barcode_set.append(barcode2yaml(bc))
    return barcode_set


def yaml2adapter(filename):
    with open(filename, 'r') as stream:
        try:
            data = yaml.load(stream)
            return data
        except yaml.YAMLError as exc:
            print(exc)


def adapter2yaml(adapter):
    return {"kit": adapter.kit,
            "auto_detect": adapter.auto_detect,
            "model": adapter.model,
            "description": adapter.description,
            "sequence": adapter.get_adapter_sequences(),
            "barcode_set_1": barcodes2yaml(adapter.get_barcode_set(0)),
            "barcode_set_2": barcodes2yaml(adapter.get_barcode_set(1))}


def read_barcode(data):
    if not data:
        return None
    return Barcode(data['name'], data['id'], data.get('sequence', None), data.get('fwd_strand', None))


def read_barcode_set(data):
    if not data:
        return None

    barcodes = []
    for barcode in data:
        barcodes.append(read_barcode(barcode))

    return barcodes


RESOURCE_PACKAGE = __name__
KIT_FOLDER = '/'.join(('resources', 'kits'))
KIT_FOLDER = pkg_resources.resource_filename(__name__, KIT_FOLDER)


def read_adapter_layout(filename):
    with open(filename, 'r') as stream:
        try:
            data = yaml.load(stream)
            if not data.get('active', True):
                return None

            model = None
            model_len = None
            if "model" in data and data['model']:
                model = data['model'].get('file', None)
                model_len = data['model'].get('length', None)

            return AdapterLayout(
                kit=data.get('kit', ""),
                sequence=data.get('sequence', ""),
                barcode_set_1=read_barcode_set(data.get('barcode_set_1', None)),
                barcode_set_2=read_barcode_set(data.get('barcode_set_2', None)),
                description=data.get('description', ""),
                auto_detect=data.get('auto_detect', False),
                model=model,
                model_len=model_len)
        except yaml.YAMLError as exc:
            print(exc)


def get_barcodes_simple(kit="standard", filename=None):
    if not filename or not os.path.isfile(filename):
        filename = os.path.join(KIT_FOLDER, "simple_{}.yml".format(kit))

    with open(filename, 'r') as stream:
        try:
            data = yaml.load(stream)
            return read_barcode_set(data.get('barcode_set_1', []))
        except yaml.YAMLError as exc:
            print(exc)


def populate_adapter_layouts(folder=None):
    if not folder:
        folder = KIT_FOLDER

    if os.path.exists(folder):
        if os.path.isdir(folder):
            filenames = glob.glob(os.path.join(folder, "*.yml"))
        else:
            filenames = [folder]
    else:
        logging.warn("{} not found. Using default adapter sequences.".format(folder))
        filenames = [KIT_FOLDER]

    adapters = []
    for filename in filenames:
        # print(filename)
        layout = read_adapter_layout(filename)
        # print(layout)
        if not layout:
            continue
        if layout.auto_detect:
            adapters.append(layout)
        else:
            adapters.append(layout)
    return adapters



try:
    import ConfigParser
except ImportError:
    import configparser as ConfigParser
import parasail


class qcatConfig:

    def __init__(self, config_path=None):

        self._match = 5
        self._nmatch = -1
        self._mismatch = -2
        self._gap_open = 2
        self._gap_extend = 2

        self._max_align_length = 150

        self._extracted_barcode_extension = 11
        self._barcode_context_length = 11

        self._matrix = None
        self.update_matrix()

        self._matrix_barcode = parasail.matrix_create("ATGCN", 1, -1)

        if config_path is not None:
            self.read(config_path)

    @property
    def matrix_barcode(self):
        """
        Scoring matrix used for aligning barcodes

        :return: parasail scoring matrix
        """
        return self._matrix_barcode


    @property
    def match(self):
        """
        Match score for adapter/barcode alignments

        :return: float
        """
        return self._match

    @match.setter
    def match(self, value):
        """
        Match score for adapter/barcode alignments

        :param value: float
        :return: None
        """
        self._match = abs(value)
        self.update_matrix()

    @property
    def nmatch(self):
        """
        Match score for adapter/barcode alignments

        :return: float
        """
        return self._nmatch

    @nmatch.setter
    def nmatch(self, value):
        """
        Match score for adapter/barcode alignments

        :param value: float
        :return: None
        """
        self._nmatch = abs(value)
        self.update_matrix()

    @property
    def mismatch(self):
        """
        Mismatch score for adapter/barcode alignments

        :return: float
        """
        return self._mismatch

    @mismatch.setter
    def mismatch(self, value):
        """
        Mismatch score for adapter/barcode alignments

        :param value: float
        :return: None
        """
        self._mismatch = -1 * abs(value)
        self.update_matrix()

    @property
    def gap_open(self):
        """
        Gap open for adapter/barcode alignments

        :return: float
        """
        return self._gap_open

    @gap_open.setter
    def gap_open(self, value):
        """
        Gap open for adapter/barcode alignments

        :param value: float
        :return: None
        """
        self._gap_open = abs(value)

    @property
    def gap_extend(self):
        """
        Gap extend for adapter/barcode alignments

        :return: float
        """
        return self._gap_extend

    @gap_extend.setter
    def gap_extend(self, value):
        """
        Gap extend for adapter/barcode alignments

        :param value: float
        :return: None
        """
        self._gap_extend = abs(value)

    @property
    def max_align_length(self):
        """
        Number of bp of the read that will be scanned for an adapter

        :return: int
        """
        return self._max_align_length

    @max_align_length.setter
    def max_align_length(self, value):
        """
        Number of bp of the read that will be scanned for an adapter

        :param value: int
        :return: None
        """
        self._max_align_length = value

    @property
    def extracted_barcode_extension(self):
        """
        Number of bp the detected barcode region from the read will be
        extended before aligning barcodes

        :return: int
        """
        return self._extracted_barcode_extension

    @extracted_barcode_extension.setter
    def extracted_barcode_extension(self, value):
        """
        Number of bp the detected barcode region from the read will be
        extended before aligning barcodes

        :param value: int
        :return: None
        """
        self._extracted_barcode_extension = value

    @property
    def barcode_context_length(self):
        """
        Number of bp upstream and downstream of the barcoded t
        hat will be added to the barcode sequence
        when aligning to the read.

        :return: int
        """
        return self._barcode_context_length

    @barcode_context_length.setter
    def barcode_context_length(self, value):
        """
        Number of bp upstream and downstream of the barcoded t
        hat will be added to the barcode sequence
        when aligning to the read.

        :param value: int
        :return: None
        """
        self._barcode_context_length = value

    def write(self, out_config_path):
        """
        Write to ini file

        :param out_config_path: ini file
        :return: None
        """
        default_config = ConfigParser.RawConfigParser()
        default_config.add_section('qcat')
        default_config.set('qcat',
                           'gap_open',
                           self.gap_open)
        default_config.set('qcat',
                           'gap_extend',
                           self.gap_extend)
        default_config.set('qcat',
                           'match',
                           self.match)
        default_config.set('qcat',
                           'mismatch',
                           self.mismatch)
        default_config.set('qcat',
                           'max_align_length',
                           self.max_align_length)
        default_config.set('qcat',
                           'extracted_barcode_extension',
                           self.extracted_barcode_extension)
        default_config.set('qcat',
                           'barcode_context_length',
                           self.barcode_context_length)

        with open(out_config_path, 'wb') as configfile:
            default_config.write(configfile)

    def update_matrix(self):
        """
        Create new parasail scoring matrix. 'N' is used as wildcard character
        for barcodes and has its own match parameter (0 per default).
        'X' is used as wildcard character for modified bp as in the 16S
        sequencing adapter.

        :return: None
        """
        self.matrix = parasail.matrix_create("ATGCNX", self.match, self.mismatch)

        pointers = [4, 11, 18, 25, 28, 29, 30, 31, 32]
        for i in pointers:
            self.matrix.pointer[0].matrix[i] = self.nmatch

        pointers = [5, 12, 19, 26, 33, 35, 36, 37, 38, 39, 40]
        for i in pointers:
            self.matrix.pointer[0].matrix[i] = 0

    def read(self, config_path):
        """
        Read from ini file

        :param config_path: ini path
        :type config_path: str
        :return: None
        """
        config = ConfigParser.RawConfigParser()
        config.read(config_path)

        self.gap_open = config.getint('qcat', 'gap_open')
        self.gap_extend = config.getint('qcat', 'gap_extend')
        self.match = config.getint('qcat', 'match')
        self.mismatch = config.getint('qcat', 'mismatch')

        self.max_align_length = config.getint('qcat',
                                              'max_align_length')
        self.extracted_barcode_extension = config.getint('qcat',
                                                         'extracted_barcode_extension')
        self.barcode_context_length = config.getint('qcat',
                                                    'barcode_context_length')


def get_default_config():
    """
    Return default config object which can be modified and passed to
    detect_barcode_* functions

    :return: Config object
    :rtype CuecCatConfig
    """
    return qcatConfig()



from __future__ import print_function

import logging

from qcat import adapters
from qcat.scanner_base import BarcodeScanner
try:
    from qcat.scanner_guppy import BarcodeScannerGuppy, guppy_import_failed
    use_pyguppy = not guppy_import_failed
except ImportError as e:
    use_pyguppy = False

from qcat.scanner_epi2me import BarcodeScannerEPI2ME
from qcat.scanner_simple import BarcodeScannerSimple
from qcat.scanner_dual import BarcodeScannerDual


def get_adapter_by_name(kit, kit_folder=None):
    """
    Get Adapter object for Kit by name

    :param kit: Name of the Kit as given in the config file
    :param kit_folder: Folder containing config files for Adapters/Kits
    :return: List of Adapter objects with given name
    """
    adapter_list = []
    for adapter in adapters.populate_adapter_layouts(kit_folder):
        if adapter.kit == kit:
            adapter_list.append(adapter)
    return adapter_list


def get_modes():
    """
    Get all available algorithms for demultiplexing.
    Currently: qcat, brill, guppy, (simple)

    :return: List of available qcat modes
    """
    modes = []
    for subclass in BarcodeScanner.__subclasses__():
        name = subclass.get_name()
        if name != "brill" or use_brill or (name == "guppy" and use_pyguppy):
            modes.append(name)
    return modes


def get_kits_info(kit_folder=None):
    """
    Get name and description of all supported kits

    :param kit_folder: Folder containing config files for Adapters/Kits
    :return: Folder containing config files for Adapters/Kits
    """
    names = {'Auto': "Auto detect kit"}
    for layout in adapters.populate_adapter_layouts(kit_folder):
        if layout.kit not in names.keys():
            names[layout.kit] = layout.description

    return names


def get_kits(kit_folder=None):
    """
    Get names (as given in the config files) for all available kits.

    :param kit_folder: Folder containing config files for Adapters/Kits
    :return: List of kit names
    """
    names = ['Auto']
    for layout in adapters.populate_adapter_layouts(kit_folder):
        if layout.kit not in names:
            names.append(layout.kit)

    return names


def factory(mode="epi2me", min_quality=None, kit=None, kit_folder=None,
            enable_filter_barcodes=False, scan_middle_adapter=False, threads=1):
    """
    Create a BarcodeScanner object

    :param mode: Algorithm to use (qcat, brill, guppy, ...)
    :param min_quality: Minimum quality to call a barcode
    :param kit: Name of the kit
    :param kit_folder: Folder containing config files for Adapters/Kits
    :param enable_filter_barcodes: Remove barcodes that occur in small numbers
    when running in batch mode
    :param scan_middle_adapter: Scan full read for adapters
    :return: BarcodeScanner object
    """

    if mode == "guppy" and not use_pyguppy:
        logging.warning(
            "Demultiplexing mode {} currently not supported in your "
            "environment. Falling back to epi2me.".format(
                mode))
        mode = "epi2me"


    for subclass in BarcodeScanner.__subclasses__():
        if mode == subclass.get_name():
            return subclass(min_quality=min_quality,
                            kit_folder=kit_folder,
                            kit=kit,
                            enable_filter_barcodes=enable_filter_barcodes,
                            scan_middle_adapter=scan_middle_adapter,
                            threads=threads
                            )

    raise RuntimeError("Invalid demultiplexing mode: {}".format(mode))


# def detect_calibration_strand(read_sequence, is_dna):
#     """
#     Legacy function for running calibration strand detection
#
#     :param read_sequence: Sequence of the read
#     :param is_dna: DNA or RNA
#     :return: Calibration strand detected true/false
#     """
#     return calibration.detect_calibration_strand(read_sequence, is_dna)


# import pkg_resources
# # import mappy
#
# RESOURCE_PACKAGE = __name__
# LAMBDA_DNA_MM2_IDX = '/'.join(('resources', 'lambda_3.6kb.fasta.minimap2.idx'))
# LAMBDA_DNA_MM2_IDX = pkg_resources.resource_filename(__name__, LAMBDA_DNA_MM2_IDX)
#
#
# def detect_calibration_strand(read_sequence, is_dna):
#     """
#     Checks whether the read_sequence stems from the lambda calibration strand
#     :param read_sequence: sequence of the read
#     :param is_dna: DNA or RNA read
#     :return: dict
#     """
#
#     if is_dna:
#         if detect_calibration_strand.minimap2_lambda_dna is None:
#             detect_calibration_strand.minimap2_lambda_dna = \
#                 mappy.Aligner(LAMBDA_DNA_MM2_IDX)
#         minimap2 = detect_calibration_strand.minimap2_lambda_dna
#     else:
#         raise NotImplementedError("RNA calibration strand detection not"
#                                   "implemented yet.")
#
#     mapped = False
#     r_len = len(read_sequence)
#     q_st = -1
#     q_en = -1
#     hits = minimap2.map(read_sequence)
#     for hit in hits:
#         if hit.is_primary:
#             q_st = hit.q_st
#             q_en = hit.q_en
#             qry_cov = (hit.q_en - hit.q_st) * 100.0 / r_len
#             if qry_cov > 90.0:
#                 mapped = True
#
#     return_dict = {"calibration_strand": mapped,
#                    "qry_start": q_st,
#                    "qry_end": q_en,
#                    "exit_status": 0
#                   }
#     return return_dict
#
#
# detect_calibration_strand.minimap2_lambda_dna = None



from __future__ import print_function

import logging
import time
import sys
import os

from six.moves import zip
from argparse import ArgumentParser, RawDescriptionHelpFormatter

from qcat import cli


def parse_args(argv):
    """
    Commandline parser

    :param argv: Command line arguments
    :type argv: List
    :param defaults: qcat config object that holds default values passed to
    ArgumentParser
    :type defaults: qcatConfig
    :return: None
    """
    usage = "Command line interface to barcode detection library"
    parser = ArgumentParser(description=usage,
                            formatter_class=RawDescriptionHelpFormatter)
    parser.add_argument(dest="FASTQ", nargs="?", default=None)
    parser.add_argument("-t", "--tsv", dest="TSV", default=None, required=False)
    parser.add_argument("-n", "--name", dest="NAME", default="-", required=False)
    parser.add_argument("-d", "--dataset", dest="DATASET", default="-", required=False)
    parser.add_argument("-s", "--summary", action="store_true", dest="SUMMARY", required=False)
    parser.add_argument("-g", "--guppy_summary", dest="GUPPY", default=None, required=False)
    parser.add_argument("-i", "--get-incorrect", dest="INCORRECT", default=None, required=False)
    args = parser.parse_args(argv)

    return args


def _parse_reads_info(comment):
    """
    Parses ONT FASTQ header information

    :param comment: FASTQ comment
    :return: dict
    """
    single_read_info = {}
    if comment:
        for pair in comment.split(" "):
            cols = pair.split('=')
            if len(cols) == 2:
                single_read_info[cols[0]] = cols[1]
    return single_read_info


def from_guppy(fastq):
    with open(fastq) as fh:

        for line in fh:
            cols = line.split("\t")
            if len(cols) > 3:
                yield cols[1].split(','), cols[2]


def get_col_number(line, pattern):
    cols = line.split('\t')
    for i, name in enumerate(cols):
        if name.strip().lower() == pattern.strip().lower():
            return i
    return None


def from_tsv(fastq):
    if fastq == '-':
        fastq = "/dev/stdin"
    with open(fastq) as fh:
        for line in fh:
            if line.startswith('name'):
                comment_col = get_col_number(line, 'comment')
                name_col = get_col_number(line, 'name')
                barcode_col = get_col_number(line, 'barcode')
                score_col = get_col_number(line, 'score')
                continue
            cols = line.strip().split("\t")
            if len(cols) > 3:
                info = _parse_reads_info(cols[comment_col])
                truebc = info['truebc'].split(",")
                yield cols[name_col], truebc, cols[barcode_col], float(cols[score_col])


def from_stdin(fastq, guppy_calls=None):
    for names, comments, sequences, qualites in cli.iter_fastx(fastq, True, 1):

        name, comment, sequence, quality = next(zip(names, comments, sequences, qualites))

        info = _parse_reads_info(comment)
        truebc = info['truebc'].split(",")
        bc = "none"
        if guppy_calls and name in guppy_calls:
            bc, score = guppy_calls[name]
            if bc == "unclassified":
                bc = "none"
        else:
            if info['barcode']:
                bc = info['barcode']
            score = -1

        yield name, truebc, bc, score


def main(argv=sys.argv[1:]):
    """
    Basic command line interface to qcat.

    :param argv: Command line arguments
    :type argv: list
    :return: None
    :rtype: NoneType
    """
    args = parse_args(argv=argv)

    correct = 0
    incorrect = 0
    fn = 0
    fp = 0

    summary = args.SUMMARY
    if not summary:
        print("program", "dataset", "read_id", "truebc", "bc", "result", "score", sep='\t')

    incorrect_only = args.INCORRECT

    guppy_calls = {}
    if args.GUPPY:
        with open(args.GUPPY) as fh:
            fh.readline()
            for line in fh:
                cols = line.split("\t")
                guppy_calls[cols[0]] = (cols[1].replace("barcode0", "").replace("barcode", ""), cols[5])

    if args.TSV:
        generator = from_tsv(args.TSV)
    else:
        generator = from_stdin(args.FASTQ, guppy_calls)

    for id, truebc, bc, score in generator:
        if bc in truebc:
            correct += 1
            if not summary:
                print(args.NAME, args.DATASET, id, ",".join(truebc), bc, "CORRECT", score, sep='\t')
        else:
            if bc == "none":
                if truebc != ["none"]:
                    fn += 1
                    if not summary:
                        print(args.NAME, args.DATASET, id, ",".join(truebc), bc, "FN", score, sep='\t')
            else:
                if truebc == ["none"]:
                    fp += 1
                    if not summary:
                        print(args.NAME, args.DATASET, id, ",".join(truebc), bc, "FP", score, sep='\t')
                else:
                    incorrect += 1
                    if not summary:
                        print(args.NAME, args.DATASET, id, ",".join(truebc), bc, "INCORRECT", score, sep='\t')


    total = correct + incorrect + fn + fp
    if total == 0:
        print("No reads found", file=sys.stderr)
        sys.exit(1)
    correct_perc = correct * 100.0 / total
    incorrect_perc = (incorrect + fp) * 100.0 / total
    fn_perc = fn * 100.0 / total

    if summary:
        print(args.NAME, args.DATASET, total, correct, incorrect, fn, fp, correct_perc,
                  incorrect_perc, fn_perc, sep="\t")


if __name__ == '__main__':

    main()



from __future__ import print_function

import sys
import logging

from qcat import config
from qcat.adapters import read_barcode
from qcat.layout import AdapterLayout
from qcat.scanner import BarcodeScanner

try:
    from pyguppy import barcoding
    guppy_import_failed = False
except ImportError as e:
    # logging.warning("Could not load pyguppy, skipping.")
    guppy_import_failed = True
    pass

from qcat.scanner_base import build_return_dict


class BarcodeScannerGuppy(BarcodeScanner):

    # This is not very nice. Should be part of the config files.
    KIT_MAPPING = {'NBD10X': 'NB',
                   'PBC096': 'BC',
                   'PBC001': 'BC',
                   'RBK001': 'RBK',
                   'RBK004': 'RBK4',
                   'LWB001': 'LWB',
                   'RPB004/RLB001': 'RLB',
                   'RAB201': 'RAB'}

    KIT_MAPPING_REV = {'NB':   'NBD10X',
                       'BC':   'PBC096',
                       'RBK':  'RBK001',
                       'RBK4': 'RBK004',
                       'LBW':  'LWB001',
                       'RLB':  'RPB004/RLB001',
                       'RAB':   'RAB201'}

    def __init__(self, min_quality=None, kit_folder=None, kit=None,
                 enable_filter_barcodes=False, scan_middle_adapter=False, threads=1):

        if min_quality is None:
            min_quality = 60

        if scan_middle_adapter:
            logging.warning("Guppy/Albacore do not support scanning for "
                            "middle adapters. Ignoring parameter.")

        super(BarcodeScannerGuppy, self).__init__(min_quality,
                                                  kit,
                                                  kit_folder=kit_folder)

        # Determine if user specified a kit
        guppy_kits = set()
        for layout in self.layouts:
            guppy_kits.update([layout.kit])

        kit_name = None
        if len(guppy_kits) == 1:
            kit_name = BarcodeScannerGuppy.KIT_MAPPING.get(guppy_kits.pop())

        if kit_name:
            # print("Kit: {}".format(kit_name), file=sys.stderr)
            self.barcoder = barcoding.Barcoder(kit_name=kit_name,
                                               num_threads=threads,
                                               min_quality=min_quality)
        else:
            # print("Auto detecting kit", file=sys.stderr)
            self.barcoder = barcoding.Barcoder(num_threads=threads,
                                               min_quality=min_quality)

    @staticmethod
    def get_name():
        return "guppy"

    @staticmethod
    def convert_guppy_result(guppy_result):

        guppy_barcode = guppy_result['barcode']
        if guppy_barcode['name'] == 'unclassified':
            guppy_barcode = None
            guppy_adapter = None
        else:
            guppy_barcode = read_barcode(guppy_barcode)
            guppy_adapter = AdapterLayout(
                kit=BarcodeScannerGuppy.KIT_MAPPING_REV.get(guppy_result['kit'],
                                                            guppy_result['kit']),
                sequence="ATGC",
                barcode_set_1=None,
                barcode_set_2=None,
                description="")

        result = build_return_dict(
            best_barcode=guppy_barcode,
            best_barcode_score=guppy_result['score'],
            best_adapter=guppy_adapter,
            best_adapter_end=90,
            trim5p=guppy_result['trim5p'],
            trim3p=guppy_result['trim3p'],
            exit_status=guppy_result['exit_status'],
        )

        return result

    def detect_barcode(self,
                       read_sequence,
                       read_qualities=None,
                       qcat_config=config.qcatConfig()):

        guppy_result = self.barcoder.detect_barcode(("", read_sequence))[0]

        return self.convert_guppy_result(guppy_result)

    def detect_barcode_batch(self, read_sequences, read_qualities=[None],
                             qcat_config=config.qcatConfig()):

        guppy_results = self.barcoder.detect_barcode_batch([("", seq, "") for seq in read_sequences])

        results = []
        for guppy_result in guppy_results:
            results.append(self.convert_guppy_result(guppy_result))

        return results



"""
Qcat setup script

Copyright (c) 2018 by Oxford Nanopore Technologies Ltd.
"""
from setuptools import setup

from qcat import __version__

setup(
    name='qcat',
    version=__version__,
    url='https://github.com/nanoporetech/qcat/',
    license='Mozilla Public License Version 2.0',
    description="Qcat is Python command-line tool for demultiplexing Oxford Nanopore reads from FASTQ files",
    author='Oxford Nanopore Technologies Ltd.',
    author_email='philipp.rescheneder@nanoporetech.com',
    install_requires=[
        'biopython',
        'parasail',
        'six',
        'pyyaml'
    ],
    extras_require={
        'DOCS': ['sphinx>=1.3.1'],
    },
    tests_require=[
        'pytest-runner',
        'pytest',
        'pytest-cov',
    ],
    packages=[
        'qcat'
    ],
    package_data={
        'qcat': [
            'resources/kits/*',
        ]
    },

    entry_points={"console_scripts": ['qcat = qcat.cli:main',
                                      'qcat-eval = qcat.eval:main',
                                      'qcat-roc = qcat.eval_roc:main',
                                      'qcat-eval-truth = qcat.eval_full:main']}
)


import logging

from qcat import config
from qcat.scanner import BarcodeScanner
from qcat.scanner_base import find_best_adapter_template, \
    extract_barcode_region, find_highest_scoring_barcode, build_return_dict


class BarcodeScannerEPI2ME(BarcodeScanner):

    def __init__(self, min_quality=None, kit_folder=None, kit=None, enable_filter_barcodes=False, scan_middle_adapter=False, threads=1):

        if min_quality is None:
            min_quality = 58

        if threads != 1:
            logging.warning("Multi threading is not yet supported in "
                            "epi2mme mode. Falling back to using a "
                            "single thread.")

        super(BarcodeScannerEPI2ME, self).__init__(min_quality,
                                                   kit,
                                                   kit_folder=kit_folder,
                                                   enable_filter_barcodes=enable_filter_barcodes,
                                                   scan_middle_adapter=scan_middle_adapter,
                                                   )
        self.barcodes = None

    @staticmethod
    def get_name():
        return "epi2me"

    def scan(self,
             read_sequence,
             read_qualities,
             bc_adapter_templates,
             nobc_adapter_templates,
             qcat_config=config.qcatConfig()):
        """
            Detects sequencing adapter and identifies best matching barcode

            :param read_sequence: Read sequence containing adapter
            :type read_sequence: str
            :param read_qualities: Base qualities of the read in Sanger encoding
            :type read_qualities: str
            :param adapter_templates: List of AdapterLayout objects
            with Ns
            :type adapter_templates: List
            :param qcat_config: qcatConfig object
            :type qcat_config: qcatConfig
            :return: see build_return_dict
            :rtype: Dictionary
            """
        exit_status = 0

        # Finding best barcoded adapters
        ret = find_best_adapter_template(adapter_templates=bc_adapter_templates,
                                         read_sequence=read_sequence,
                                         qcat_config=qcat_config)

        best_adapter_template_index, aligned_adapter_end, best_adapter_score = ret

        # if barcoded adapter found
        best_adapter_template = bc_adapter_templates[
            best_adapter_template_index]

        # Detect barcode
        barcode_set_index = 0

        # For high quality adapter alignments just use the barcode region,
        # for low quality compare full adapter to the barcodes
        # Do not scann full adapter is it contains double barcoding
        # best_adapter_score > 90.0 or
        if best_adapter_score > 90.0 or best_adapter_template.is_doulbe_barcode():
            barcode_region_read = extract_barcode_region(
                read_sequence=read_sequence,
                adapter_template=best_adapter_template,
                barcode_set_index=barcode_set_index,
                alignment_stop_ref=aligned_adapter_end,
                qcat_config=qcat_config)
        else:
            barcode_region_read = read_sequence[:qcat_config.max_align_length]

        # First barcode
        up_context = best_adapter_template.get_upstream_context(
            qcat_config.barcode_context_length, barcode_set_index)
        down_context = best_adapter_template.get_downstream_context(
            qcat_config.barcode_context_length, barcode_set_index)
        barcode_set = best_adapter_template.get_barcode_set(
            barcode_set_index)
        if self.barcodes:
            barcode_set = self.barcodes

        # Find best barcode
        best_barcode, best_barcode_q_score, best_barcode_score, barcode_end = \
            find_highest_scoring_barcode(
                barcode_region_read=barcode_region_read,
                barcode_set=barcode_set,
                upstream_context=up_context,
                downstream_context=down_context,
                qcat_config=qcat_config)

        # If double barcode adapter
        barcode_set_index = 1
        if best_adapter_template.get_barcode_set(barcode_set_index):

            barcode_region_read = extract_barcode_region(
                read_sequence=read_sequence,
                adapter_template=best_adapter_template,
                barcode_set_index=barcode_set_index,
                alignment_stop_ref=aligned_adapter_end,
                qcat_config=qcat_config)

            # First barcode
            up_context = best_adapter_template.get_upstream_context(
                qcat_config.barcode_context_length, barcode_set_index)
            down_context = best_adapter_template.get_downstream_context(
                qcat_config.barcode_context_length, barcode_set_index)
            barcode_set = best_adapter_template.get_barcode_set(
                barcode_set_index)
            if self.barcodes:
                barcode_set = self.barcodes

            # Find best barcode
            best_barcode_2, best_barcode_q_score_2, best_barcode_score_2, barcode_end_2 = \
                find_highest_scoring_barcode(
                    barcode_region_read=barcode_region_read,
                    barcode_set=barcode_set,
                    upstream_context=up_context,
                    downstream_context=down_context,
                    qcat_config=qcat_config)
        else:
            logging.debug("Adapter type does not have second barcode")

        return build_return_dict(
            best_barcode=best_barcode,
            best_barcode_score=best_barcode_q_score,
            best_adapter=best_adapter_template,
            best_adapter_end=aligned_adapter_end,
            exit_status=exit_status
        )
